/*   ----------------- AVRTETRIS.C -----------------
     Tetris fuer den ATmega8 in Verbindung
     mit einem Nokia 3410 Display
     
     IDE-Platform:  AVR Studio4
     01.01.2013

     ----------------------------------------------- */

#include <util/delay.h>
#include <avr/io.h>
#include "graph3410.h"
#include "tetlogo0.h"


/* Belegung Tastenpins:
                   PD2 = Taster links
                   PD3 = Taster rechts
                   PD4 = Taster start */

// ----------------- "Tastenbit-Macros" ------------------

#define tast_port   PORTD
#define tast_in     PIND
#define tast_ioreg  DDRD
#define tastl 2                  // PortD.2
#define tastr 3                  // PortD.3
#define tasts 4                  // PortD.4
#define tastd 5                  // PortD.5
#define tast_inmask        ((1 << tastl)|(1 << tastr)|(1 << tasts)|(1 << tastd))

#define isbitD(nr)         ((PIND) & (1<<nr))
#define intbit(wert,nr)    ((wert) & (1<<nr))
#define setintbit(wert,nr) (wert |= (1<<nr))
#define clrintbit(wert,nr) (wert &= ~(1<<nr))

#define klotzsize 5
#define rasterxoffset 5
#define rasteryoffset 0
#define figanz 7
#define scorepos           11,0
#define intime             60
#define outerloop          2

const uint16_t tfig01 [figanz][4] = {{0xe400,0x4c40,0x4e00,0x4640},
                                     {0xe200,0x44c0,0x8e00,0x6440},
                                     {0xe800,0xc440,0x2e00,0x4460},
                                     {0x6c00,0x8c40,0x6c00,0x8c40},
                                     {0xc600,0x4c80,0xc600,0x4c80},
                                     {0xcc00,0xcc00,0xcc00,0xcc00},
                                     {0xf000,0x4444,0xf000,0x4444}};

uint16_t raster [14];
uint8_t istr_buffer [7];
char aktfig, aktrot;
char figx, figy;
uint8_t hb;                          // Hilfsbyte
uint8_t loopcx;                      // Schleifenzaehler fuer Zeit und Tastenaktion
uint8_t innerloop;
char rnd;                            // wird kontinuierlich hochgezaehlt (Pseudozufall)
char action;                         // Auszufuehrende Aktion waehrend des Spiels
uint16_t hint;                       // Hilfsinteger
uint16_t scorecx;
uint8_t gameover;


void initraster(void)
{
  char i;
  for (i= 0;i<13;i++) {raster[i]= 0x8010;}
  raster[13]= 0xffff;
}

void delay(char wert)
{
   char i;
   for (i=0;i< wert;i++)
   {
     _delay_ms(100);
   }
}

void tasten_init()
{
  tast_ioreg &= ~(tast_inmask);    // Pins als Eingang schalten
  tast_port |= tast_inmask;        // interne Pull-Up Widerstnde einschalten
}

void drawklotz(char xpos, char ypos, char mode)
/* ein Kloetzchen ist <klotzsize> x Pixel gross.
   Ein Display ist in dieses Raster unterteilt. Die
   Offsetkoordinaten geben die Position in einem solchen
   Raster an !   
   mode = 1 ==> Klotz wird gezeichnet
   mode = 0 ==> Klotz wird geloescht */
{
  char x,y;

   for (y= 0; y < klotzsize-1; y++)
   {
     for (x= 0; x < klotzsize-1; x++)
     {
       putpixel((xpos*klotzsize)+rasterxoffset+x,(ypos*klotzsize)+rasteryoffset+y,mode);
     }
   }
}

void drawfig(char xpos, char ypos, uint16_t figur, char mode)
/* zeichnet eine Figur an die angegebene Position <xpos>,<ypos>
   unter Beruecksichtigung der Offsetkoordinaten */
{
  char x,y,i;
  
  for (i= 15; i> -1; i--)
  {
    x= (15-i)%4;
    y= (15-i)/4;
    if (intbit(figur,i)) drawklotz(x+xpos,y+ypos,mode);
  }
}


void showscore(void)
{
  gotoxy(scorepos); 
  LcdCacheIdx = (163);                                  // Ausgabe des Scores ausserhalb des Textkoordinatenrasters

  itoa(scorecx,istr_buffer,10);
  if (scorecx< 10) {putlcdchar('0');}
  if (scorecx< 100) {putlcdchar('0');}
  if (scorecx< 1000) {putlcdchar('0');}
  putramstring(istr_buffer);
}

void drawraster(void)
/* zeichnet das Spielfeld neu */
{
  char x,y;
  
  for (y= 0;y< 13;y++)
  {
    for (x= 1; x< 11; x++)
    {
      if (intbit(raster[y],15-x)) {drawklotz(x,y,1);} else {drawklotz(x,y,0);}
    }
  }
  // line(rasterxoffset+2,0,rasterxoffset+2,63,1);                            // Begrenzungslinien
  //line(rasterxoffset+56,0,rasterxoffset+56,63,1);
  showimage(0,0,&gamscrl[0][0],gamscrlx,gamscrly,1);    showscore();
}

char canposxy(char fx, char fy, uint16_t fig)
/* testet, ob an der angegebenen Position die uebergebene Figur
   positionierbar ist ! */

{
  char x,y,ind;
  ind= 15;
  for (y= fy; y< (fy+4); y++)
  {
    for (x= fx; x< (fx+4); x++)
    {
      if ((intbit(raster[y],15-x)>0) & (intbit(fig,ind)>0)) {return 0;}
      ind--;
    }
  }
  return 1;
}

void copyfig(void)
/* kopiert die aktuelle Spielefigur in der aktuellen Rotationslage
   ins Spieleraster */
{
  char x,y,ind;
  uint16_t fig;

  ind= 15;
  fig= tfig01[aktfig][aktrot];
  for (y= figy; y< (figy+4); y++)
  {
    for (x= figx; x< (figx+4); x++)
    {
      if (intbit(fig,ind)>0) {setintbit(raster[y],15-x);}
      ind--;
    }
  }
}

void scanraster(void)
/* Scant das Spielfeld auf vollstaendig ausgefuellte Reihen und
   entfernt eine solchige bei gleichzeitiger Erhoehunge des
   Spielstands <scorecx> um 2 */
{
  char i,i2;
  for (i= 12; i> -1; i--)
  {
    while ((raster[i]&0xfff0)==0xfff0)
    {
      scorecx += 3;
      if (i==0) {raster[0]= 0x8010;}
        else for (i2= i; i2> 0; i2--) {raster[i2]= raster[i2-1];}
    }
  }
}

void checkdownrow(void)
{
  copyfig();
  delay(4);
  scanraster();
  drawraster();
  delay(4);
  aktfig= rnd; aktrot= 0; figx= 5; figy= 0;
  action= 0;
  if (canposxy(figx,figy,tfig01[aktfig][aktrot]))
  {
    drawfig(figx,figy,tfig01[aktfig][aktrot],1);
  }
  else {gameover= 1;}
  scr_update();
}


/* --------------------------------------------------------
                        M  A  I  N
   -------------------------------------------------------- */

int main(void){
  unsigned char b,f;
  
  uint16_t test;
  test= 0;

  spi_init();
  tasten_init();
  initraster();
  delay(1);
  init_lcd();
  showimage(0,0,&tetlogo[0][0],tetlogox,tetlogoy,1);                // "Eingangsbildschirm" anzeigen
  scr_update();

  while(isbitD(tasts));                                             // Startscreen solange anzeigen bis Taste gedrueckt ist
  delay(1);
  while(!isbitD(tasts));
  delay(2);
  while(1)
  {
    clrscr();
    scorecx= 0;
    scanraster();
    drawraster();
    /*for (hb= 0; hb<8; hb++)
    {
      gotoxy(0,hb); putlcdchar(92);
    }  */

    showimage(61,0,&gamscr[0][0],gamscrx,gamscry,1);

    aktfig= 0; aktrot= 0; figx= 5; figy= 0;
    action= 0;
    loopcx= outerloop;
    drawfig(figx,figy,tfig01[aktfig][aktrot],1);
    scr_update();
    gameover= 0;
    while(gameover==0)
    {
      switch (action)
      {
        case 20:                                                      // Testzweck : Figur auf Raster uebernehmen
          clrscr();
          scr_update();
          delay(7);
          copyfig();
          drawraster();
          scr_update();
          delay(7);
          aktfig= 0; aktrot= 0;
          clrscr();
          drawraster();
          figx= 5; figy= 2;
          drawfig(figx,figy,tfig01[aktfig][aktrot],1);
          scr_update();
          break;

        case 1:                                                       // rotieren
          hb= aktrot;
          aktrot++;
          aktrot %= 4;
          hint= tfig01[aktfig][aktrot];
          if (aktfig==6) {hint= 0xffff;}                              // Figur 7 (langer Stab) kann nur
                                                                    // gedreht werden, wenn alle Plaetze
                                                                    // unter ihm unbelegt sind !!!
          if (canposxy(figx,figy,hint))
          {
            drawfig(figx,figy,tfig01[aktfig][hb],0);
            drawfig(figx,figy,tfig01[aktfig][aktrot],1);
            scr_update();
          }
          else
          {
            aktrot= hb;                                               // rotieren war nicht moeglich;
          }
          break;

        case 2:                                                       // rechts schieben
          if (canposxy(figx+1,figy,tfig01[aktfig][aktrot]))
          {
            drawfig(figx,figy,tfig01[aktfig][aktrot],0);
            figx++;
            drawfig(figx,figy,tfig01[aktfig][aktrot],1);
            scr_update();
          }
          break;

        case 3:                                                       // links schieben
          if (canposxy(figx-1,figy,tfig01[aktfig][aktrot]))
          {
            drawfig(figx,figy,tfig01[aktfig][aktrot],0);
            figx--;
            drawfig(figx,figy,tfig01[aktfig][aktrot],1);
            scr_update();
          }
          break;
        case 4:
          while (canposxy(figx,figy+1,tfig01[aktfig][aktrot])) {figy++;}
          scorecx++;
          checkdownrow();
          break;
      }


      if (innerloop> 0)
      {
        for (hb= innerloop; hb> 0; hb--)
        {
          _delay_ms(4);
        }
      }

      action= 0;
      while ((action==0) & (loopcx>0))
      {
        loopcx--;
        rnd++;
        rnd %= figanz;
        innerloop= intime;
        while ((action==0)&(innerloop>0))
        {
          if (!isbitD(tasts)) {action= 1;}
          if (!isbitD(tastr)) {action= 2;}
          if (!isbitD(tastl)) {action= 3;}
          if (!isbitD(tastd)) {action= 4;}
          _delay_ms(2);
          innerloop--;
        }
        if ((action<4)&(action>0)) {innerloop= intime;}
      }
      if (loopcx== 0)
      {
        loopcx= outerloop;
        if (canposxy(figx,figy+1,tfig01[aktfig][aktrot]))
        {
          drawfig(figx,figy,tfig01[aktfig][aktrot],0);
          figy++;
          drawfig(figx,figy,tfig01[aktfig][aktrot],1);
          scr_update();
        }
        else
        {
          scorecx++;
          checkdownrow();
        }
      }
    }

    for (hb= 21; hb< 31+gologoy; hb++)
    {
      line(11,hb,21+(gologox*8),hb,1);
    }
    rectangle(12,22,20+(gologox*8),29+gologoy,0);
    showimage(16,26,&gologo[0][0],gologox,gologoy,0);                // "Eingangsbildschirm" anzeigen
    scr_update();
    while(isbitD(tastd));                                            // Taste D fuer einen Neustart
    delay(1);
    while(!isbitD(tastd));
    initraster();
    delay(2);
  }
}

