#include "graph3410.h"

/* -------------------- GRAPH3410.AVC -----------------------
   Funktionen die Anbindung eines Nokia 3410 Display's
   an einen AVR-Mikrocontroller

   30.12.2012 R. Seelig
   ------------------------------------------------------------ */


/* -------------------------------------------------------------------------------------------------------------------------------
   SPI Interface
   ------------------------------------------------------------------------------------------------------------------------------- */

/* -------------------------------------------------------------
   SPI_INIT
   
   SPI-Hardwareinterface des ATmega konfigurieren
   ------------------------------------------------------------- */

void spi_init(void)
// SPI Hardware initialisieren
{
  SPI_DDR |= _BV(SPI_MOSI_PIN) | _BV(SPI_CLK_PIN) | _BV(SPI_SS_PIN);    // Enable SPI port: No interrupt, MSBit first, Master mode,
  SPCR = 0x50;
}

/* -------------------------------------------------------------
   SPI
   
      Byte ueber SPI senden / empfangen
      data ==> zu sendendes Datum
   ------------------------------------------------------------- */

unsigned char spi(unsigned char data)
{
  SPDR=data;                                // Datum senden
  while( ( SPSR & 0x80 ) != 0x80 );         // ... und warten bis es gesendet ist
  return SPDR;                              // empfangenes Byte ist Rueckgabewert
}


/* -------------------------------------------------------------------------------------------------------------------------------
   LCD Interface
   ------------------------------------------------------------------------------------------------------------------------------- */

/* -------------------------------------------------------------
   LCD_SEND

   sendet Data via SPI an das LCD
          LcdCmdData = LCD_CMD  ==> data als Kommando schicken
                     = LCD_DATA ==> data als Datum schicken
   ------------------------------------------------------------- */

void lcd_send(LcdCmdData cd, unsigned char data)

{

  if (cd == LCD_DATA) LCD_PORT |= _BV(LCD_DC_PIN);   // Display C/D setzen
  else LCD_PORT &= ~(_BV(LCD_DC_PIN));

  LCD_PORT &= ~(_BV(LCD_CE_PIN));                    // Display Controller enable
  spi(data);

  LCD_PORT |= _BV( LCD_CE_PIN );                     // Disable Controller
}

/* -------------------------------------------------------------
   INIT_LCD

   initialisert LCD
   ------------------------------------------------------------- */

void init_lcd(void)

{
  int n;

  LCD_DDR |= _BV(LCD_RST_PIN) | _BV(LCD_DC_PIN);     // Set LCD Output pins
  LCD_PORT &= ~(_BV(LCD_RST_PIN));                   // Resets LCD controler
  _delay_ms(1);
  LCD_PORT |= _BV(LCD_RST_PIN);                      // Set LCD CE = 1 (Disabled)
  LCD_PORT |= _BV(LCD_CE_PIN);

  // LCD Controller Kommandos  (eigentliches initialisieren)

  lcd_send(LCD_CMD,0x21);                            // Erweiterter Kommandomodus
  lcd_send(LCD_CMD,0x09);                            // Set int. HV Generator (ca. 7V an Pin7)
  lcd_send(LCD_CMD,0xff);                            // VOP max
  lcd_send(LCD_CMD,0x16);                            // BIAS = 2
  lcd_send(LCD_CMD,0x06);                            // Temp. Koeffizient = 2
  lcd_send(LCD_CMD,0x20);                            // Standart Kommandomodus
  lcd_send(LCD_CMD,0x0c);                            // normale Ausgabe (normal = 0Ch, invertiert = 0Dh)


  memset(LcdCache,0x00,LCD_CACHE_SIZE);              // LCD Cache loeschen

  // LCD loeschen

  lcd_send(LCD_CMD, 0x80);
  lcd_send(LCD_CMD, 0x40);
  for(n=0; n<(LCD_REAL_X_RES*LCD_REAL_Y_RES); n++) lcd_send(LCD_DATA,0x00);
  akttxpos_x= 0;
  akttxpos_y= 0;
}


/* -------------------------------------------------------------
   CLRSCR

   loescht den Screenspeicher
   ------------------------------------------------------------- */

void clrscr(void)
{
  int n;
  
  memset(LcdCache,0x00,LCD_CACHE_SIZE);              // LCD Cache loeschen
  lcd_send(LCD_CMD, 0x80);
  lcd_send(LCD_CMD, 0x40);
  for(n=0; n<(LCD_REAL_X_RES*LCD_REAL_Y_RES); n++) lcd_send(LCD_DATA,0x00);
  gotoxy(0,0);
}

/* -------------------------------------------------------------
   GOTOXY

   positioniert den Textcursor im Screenspeicher an
   Position x,y (Textkoordinaten x= 0..15 / y= 0..7 )
   ------------------------------------------------------------- */

void gotoxy(unsigned char x, unsigned char y)
{

  if(x>((LCD_VISIBLE_X_RES/6)-1)) return;               // Funktion verlassen wenn X-Koordinate ausserhalb des Displays
  if(y>((LCD_VISIBLE_Y_RES/8)-1)) return;               // dto. Y Koordinate

  LcdCacheIdx = ( x * 6) + ( y * LCD_VISIBLE_X_RES);    // Adresse im LCD-Cache berechnen
  akttxpos_x= x;
  akttxpos_y= y;
}

/* -------------------------------------------------------------
   PUTLCDCHAR

   setzt ein Zeichen im Screenspeicher und postioniert den
   Cursor auf die naechste Position
   
          ch ==> zu setzenes Zeichen
   ------------------------------------------------------------- */

void putlcdchar (unsigned char ch)
{
  unsigned char i;

  if (ch== 13)                                          // Fuer <printf> "/r" Implementation
  {
    gotoxy(0,akttxpos_y);
    return;
  }
  if (ch== 10)                                          // fuer <printf> "/n" Implementation
  {
    gotoxy(akttxpos_x,akttxpos_y+1);
    return;
  }
  if ((ch<0x20)||(ch>lastascii)) ch = 92;               // ASCII Zeichen umrechnen

  // Kopiere Daten eines Zeichens aus dem Zeichenarray in den LCD-Cache

  for (i=0; i<5; i++) LcdCache[LcdCacheIdx++]=pgm_read_byte(&(FontLookup[ch-32][i]))<<1;

  LcdCache[LcdCacheIdx]=0x00;

  if(LcdCacheIdx==(LCD_CACHE_SIZE-1))
  {
    LcdCacheIdx = 0;
    akttxpos_x= 0;
    akttxpos_y++;
  }
  else 
  {
    LcdCacheIdx++;
    akttxpos_x++;
  }
}


/* -------------------------------------------------------------
   PUTRAMSTRING

   kopiert einen String aus dem RAM (Array) in denScreenspeicher
   ab der aktuellen Cursorposition

          dataArray ==> zu setzender String (innerhalb eines 
                                             Arrays)
   ------------------------------------------------------------- */

void putramstring(unsigned char dataArray[])
{
  unsigned char tmpIdx=0;

  while(dataArray[tmpIdx]!='\0') putlcdchar(dataArray[tmpIdx++]);
}

/* -------------------------------------------------------------
   PUTROMSTRING

   kopiert einen String aus dem ROM in den Screenspeicher
   ab der aktuellen Cursorposition. 
   
          *dataPtr ==> Zeiger auf den String im ROM

   ------------------------------------------------------------- */

void putromstring(const unsigned char *dataPtr)
{
  unsigned char c;

  for (c=pgm_read_byte(dataPtr); c; ++dataPtr, c=pgm_read_byte(dataPtr)) putlcdchar(c);
}


/* -------------------------------------------------------------
   PUTPIXEL

   Setzt Pixel an Position X,Y im Screenspeicher
   PixelMode 0 = loeschen
             1 = setzen
             2 = Pixelpositon im XOR-Modus verknuepfen

   ------------------------------------------------------------- */

void putpixel(unsigned char x, unsigned char y, PixelMode mode )
{
  unsigned int index;
  unsigned char offset;
  unsigned char data;

  index=((y/8)*LCD_VISIBLE_X_RES)+x;
  offset=y-((y/8)*8);
 
 // Pixel im entsprechenden PixelMode setzen

  data=LcdCache[index];
  if (mode==PIXEL_ON) data |= (0x01<<offset);                          // Pixel setzen
  else if (mode==PIXEL_OFF) data &= (~(0x01<<offset));                 // Pixel loeschen
  else if (mode==PIXEL_XOR) data ^= (0x01<<offset);                    // Pixel im XOR-Mode setzen
  LcdCache[index]=data;
}


/* -------------------------------------------------------------
   LINE

   Zeichnet eine Linie von den Koordinaten x0,y0 zu x1,y1
   im Screenspeicher.
   
   Linienalgorithmus nach Bresenham (www.wikipedia.org)

   PixelMode 0 = loeschen
             1 = setzen
             2 = Pixelpositon im XOR-Modus verknuepfen

   ------------------------------------------------------------- */

void line(int x0, int y0, int x1, int y1, PixelMode mode)
{


  //    Linienalgorithmus nach Bresenham (www.wikipedia.org)

  int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int err = dx+dy, e2;                                     /* error value e_xy */
 
  for(;;)
  {
    putpixel(x0,y0,mode);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 > dy) { err += dy; x0 += sx; }                  /* e_xy+e_x > 0 */
    if (e2 < dx) { err += dx; y0 += sy; }                  /* e_xy+e_y < 0 */
  }
}

/* -------------------------------------------------------------
   RECTANGLE

   Zeichnet ein Rechteck von den Koordinaten x0,y0 zu x1,y1
   im Screenspeicher.
   
   Linienalgorithmus nach Bresenham (www.wikipedia.org)

   PixelMode 0 = loeschen
             1 = setzen
             2 = Pixelpositon im XOR-Modus verknuepfen

   ------------------------------------------------------------- */

void rectangle(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, PixelMode mode)
{
  line(x1,y1,x2,y1, mode);
  line(x2,y1,x2,y2, mode);
  line(x1,y2,x2,y2, mode);
  line(x1,y1,x1,y2, mode);
}

/* -------------------------------------------------------------
   ELLIPSE

   Zeichnet eine Ellipse mit Mittelpunt an der Koordinate xm,ym
   mit den Hoehen- Breitenverhaeltnis a:b
   im Screenspeicher.
   
   Ellipsenalgorithmus nach Bresenham (www.wikipedia.org)

   PixelMode 0 = loeschen
             1 = setzen
             2 = Pixelpositon im XOR-Modus verknuepfen

   ------------------------------------------------------------- */

void ellipse(int xm, int ym, int a, int b, PixelMode mode )
{
  // Algorithmus nach Bresenham (www.wikipedia.org)

  int dx = 0, dy = b;                       // im I. Quadranten von links oben nach rechts unten
  long a2 = a*a, b2 = b*b;
  long err = b2-(2*b-1)*a2, e2;             // Fehler im 1. Schritt */

  do
  {
    putpixel(xm+dx, ym+dy,mode);            // I.   Quadrant
    putpixel(xm-dx, ym+dy,mode);            // II.  Quadrant
    putpixel(xm-dx, ym-dy,mode);            // III. Quadrant
    putpixel(xm+dx, ym-dy,mode);            // IV.  Quadrant

    e2 = 2*err;
    if (e2 <  (2*dx+1)*b2) { dx++; err += (2*dx+1)*b2; }
    if (e2 > -(2*dy-1)*a2) { dy--; err -= (2*dy-1)*a2; }
  } while (dy >= 0);

  while (dx++ < a)                        // fehlerhafter Abbruch bei flachen Ellipsen (b=1)
  {
    putpixel(xm+dx, ym,mode);             // -> Spitze der Ellipse vollenden
    putpixel(xm-dx, ym,mode);
  }
}

/* -------------------------------------------------------------
   CIRCLE

   Zeichnet einen Kreis mit Mittelpunt an der Koordinate xm,ym
   und dem Radius r im Screenspeicher.
   
   PixelMode 0 = loeschen
             1 = setzen
             2 = Pixelpositon im XOR-Modus verknuepfen

   ------------------------------------------------------------- */

void circle(int x, int y, int r, PixelMode mode )
{
  ellipse(x,y,r,r,mode);
}



/* -------------------------------------------------------------
   SHOWIMAGE

   Kopiert ein im Flash abgelegtes Bitmap in den Screenspeicher.
   Bitmap muss byteweise in Zeilen gespeichert vorliegen
   Hierbei entspricht 1 Byte 8 Pixel.
   Bsp.: eine Reihe mit 6 Bytes entsprechen 48 Pixel in X-Achse
   
   ox,oy        => linke obere Ecke an der das Bitmap angezeigt wird
   image        => das anzuzeigende Bitmap
   resX         => Anzahl der Bytes in X-Achse
   LcdPixelMode => Ausgabemode
      0 = Bitmappixel loeschen
      1 = Bitmappixel setzen
      2 = Bitmappixel im XOR-Mode mit Hintergrund verknuepfen

   ------------------------------------------------------------- */

void showimage(char ox, char oy, const unsigned char* const image,uint8_t resX,uint8_t resY, PixelMode mode)
{
  int x,y;
  char b,bp;
  for (y=0;y< resY;y++)
  {
    for (x= 0;x<resX;x++)
    {
      b= pgm_read_byte(&(image[y *resX +x]));
      for (bp=8;bp>0;bp--)
      {
        if (b& 1<<bp-1) {putpixel(ox+(x*8)+8-bp,oy+y,mode);}
      }
    }
  }
}

/* -------------------------------------------------------------
   SCR_UPDATE

   bringt den Screenspeicher (abgelegt im ATmega) zur Anzeige
   auf dem LCD

   ------------------------------------------------------------- */

void scr_update(void)
{
  unsigned int i=0;
  unsigned char row;
  unsigned char col;

  for (row=0; row<(LCD_VISIBLE_Y_RES / 8); row++)
  {
    lcd_send(LCD_CMD, 0x80);
    lcd_send(LCD_CMD, 0x40 | row);
    for (col=0; col<LCD_VISIBLE_X_RES; col++) lcd_send(LCD_DATA,LcdCache[i++]);
  }
}

/* -------------------------------------------------------------
   LCD_FILEOUT

   Dateiausgabe. Ein Zeichen wird auf dem angegebenen File aus-
   gegeben!!!
   
   Wird die Standardausgabe hierhin umgeleitet sendet bspw.
   <printf> jedes einzelne Zeichen hin

   ------------------------------------------------------------- */

static int lcd_fileout(char ch, FILE *stream)
{
  putlcdchar(ch);
}

